/*
 * AlibavaSOAP.cc
 *
 *  Created on: 9 Jan 2011
 *      Author: lacasta
 */

#include "alibavaService.h"
#include "AlibavaSOAP.h"
#include <utils.h>
#include <pthread.h>

class Thread
{
    private:
        /// The thread we are wrapping
        pthread_t _thread;
        char *thread_w;

        /// says if the thread has already started
        bool           _started;

        /// says if the thread has stopped
        bool           _stopped;

        /// The nice value for the thread process
        int            _nice;
    public:
        enum CancelType {
            deferred,
            asynchronous,
            disable
        };

        /// constructor
        Thread();

        /// destructor
        virtual ~Thread()
        {};

        /// Sets the nice value of the thread process
        void nice(int x)
        {
            _nice = x;
        }

        /// test if the thread has been cancelled
        void test_cancel();

        /// sets the way the thread can be cancelled. It is any of
        /// the options in CancelType
        void set_cancel(CancelType);

        /// exits the thread
        void exit(int v = 0);

        /// Starts the thread
        virtual void start();

        /// Stops the thread
        virtual void stop();

        /// The method that gets executed in the thread.
        virtual void run() = 0;

        /// Internal
        static void *_run(void *);

        /// Checks if the thread is running
        bool is_running()
        {
            return _started;
        }

        /// Checks if the thread has stopped
        bool is_stopped()
        {
            return _stopped;
        }

};
Thread::Thread() :  thread_w(0), _started(false), _stopped(false), _nice(-1)
{
}


void Thread::start()
{
    _started = true;
    _stopped = false;

    /* Initialize and set thread detached attribute */
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    pthread_create(&_thread, &attr, _run, this);

    pthread_attr_destroy(&attr);

    thread_w = (char *)1;
}

void Thread::stop()
{
    void *rc;
    if (!_stopped && thread_w)
    {
        _stopped = true;
        pthread_cancel(_thread);
        pthread_join(_thread, &rc);
    }
    _started = false;
    _stopped = true;
    thread_w = (char *)0;
}

void Thread::exit(int val)
{
    pthread_exit((void *)val);
}

void Thread::test_cancel()
{
    pthread_testcancel();
}

void Thread::set_cancel(CancelType ct)
{
    switch (ct)
    {
        case deferred:
            pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, 0);
            pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, 0);
            break;
        case asynchronous:
            pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, 0);
            pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0);
            break;
        case disable:
            pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, 0);
            break;
    }
}

void *Thread::_run(void *data)
{
    Thread *thrd = reinterpret_cast<Thread *>(data);
    //    if (thrd->_nice>=0)
    //       setpriority(PRIO_PROCESS,getpid(),thrd->_nice);

    thrd->run();

    thrd->_stopped = true;
    thrd->_started = false;
    pthread_exit(0);
    return 0;
}


class AlibavaServer : public Thread
{
    private:
        int port;
        int _error;
        alibavaService server;

    public:
        AlibavaServer(int p) : port(p),  _error(0)  {
            ((soap *)&server)->bind_flags |= SO_REUSEADDR;
        }
        ~AlibavaServer()
        {
            close_socket();
        }
        void run();

        void close_socket()
        {
            server.soap_close_socket();
        }
        int error()
        {
            return _error;
        }

        soap *get_soap() { return (soap *)&server; }
};
void AlibavaServer::run()
{
    _error = server.run(port);
    write_msg("Quitting SOAP server thread. rc=%d\n", _error);
}

struct AlibavaSOAP_priv
{
        AlibavaServer server;
        AlibavaSOAP_priv(int port) : server(port) {}
};

AlibavaSOAP::AlibavaSOAP(int port)
{
    _priv = new AlibavaSOAP_priv(port);
}

AlibavaSOAP::~AlibavaSOAP()
{
    delete _priv;
}


void AlibavaSOAP::serve(bool threaded)
{
    if (threaded)
        _priv->server.start();
    else
        _priv->server.run();
}

int AlibavaSOAP::error()
{
    return _priv->server.error();
}

std::string AlibavaSOAP::strerror() const
{
   return std::string( *soap_faultstring(_priv->server.get_soap()) );
}

void AlibavaSOAP::stop_serving()
{
    _priv->server.close_socket();
}

bool AlibavaSOAP::is_running() const
{
    return _priv->server.is_running();
}
