/*
 * DAQmanager.cc
 *
 *  Created on: 29 Jan 2011
 *      Author: lacasta
 */

#include <iostream>
#include <sstream>
#include "utils.h"
#include "DAQmanager.h"
#include <RunManager.h>
#include <DataFile.h>
#include <DataMonitor.h>
#include <AsciiDataFile.h>
#include "ConfigFactory.h"
#include "Analysis.h"
#include "DAQConfig.h"
#include "PluginConfig.h"

DAQmanager::DAQmanager(int nn, Alibava *B, RunManager::RunType tp) :
    nevts(nn), received(0), expected(nn), board(B), ofile(0), monitor(0),
    run_type(RunManager::Unknown), rm(0),
    _starting(false), _running(false), _stop_running(false),  config_file( 0)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    config_file = new KeyFile();
    cfg->set_file(config_file);

    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));

    daq->nevts(nn);
    board->set_data_receiver(this);
    board->delay( daq->delay() );
    set_run_type(tp);
    mtx = g_mutex_new();
}

DAQmanager::~DAQmanager()
{
    g_mutex_free(mtx);
    if (rm)
        delete rm;

    delete config_file;

}


bool DAQmanager::running() const
{
    bool rc;
    g_mutex_lock(mtx);
    rc = _running;
    g_mutex_unlock(mtx);
    return rc;
}
void DAQmanager::running(bool x)
{
    g_mutex_lock(mtx);
    _running = x;
    _starting = false;
    g_mutex_unlock(mtx);
}


bool DAQmanager::starting() const
{
    bool rc;
    g_mutex_lock(mtx);
    rc =  _starting;
    g_mutex_unlock(mtx);
    return rc;
}

bool DAQmanager::stop() const
{
    bool rc;
    g_mutex_lock(mtx);
    rc =  _stop_running;
    g_mutex_unlock(mtx);
    return rc;
}
void DAQmanager::stop(bool x)
{
    g_mutex_lock(mtx);
    _stop_running = x;
    g_mutex_unlock(mtx);
}

int DAQmanager::open(const char *fnam)
{
    board->open(fnam);
    if (!(*board))
    {
        err_msg( _("Something failed when opening the device") );
        return 1;
    }

    // reset the card
    if (reset()<0)
        return -1;

    device_opened();
    return 0;
}

int DAQmanager::reset()
{
    int rc;
    write_msg(_("*** Reseting the board\n"));
    // First thing we do is to send a reset
    rc = board->software_reset();
    if (rc)
        return -1;

    return ConfigFactory::instance()->config_board(*board);
}


void DAQmanager::set_run_type( RunManager::RunType tp )
{
    if (tp == RunManager::Unknown || tp == run_type )
        return;

    if (rm)
        delete rm;

    run_type = tp;
    rm = RunManager::factory(tp, this);
    if (rm)
    {
        ConfigFactory *cfg = ConfigFactory::instance();
        PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
        if ( plg->active())
            rm->set_plugin(plg->plugin());
    }
    if (!rm)
    {
        err_msg( _("Could not create an appropriate RunManager\n") );
    }
}

void DAQmanager::set_data_file(const std::string &fnam)
{
    if (!fnam.empty())
    {
        ofile = new AsciiDataFile(fnam);
    }
}

int DAQmanager::prepare_for_run()
{
    ConfigFactory *cfg = ConfigFactory::instance();
    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));


    if (!rm)
        return -1;

    received = 0;
    _starting = true;
    board->stop(true);
    board->delay( daq->delay() );
    ana->reset(true);
    /*
     * Prepare the output file (if any)
     */
    if (ofile)
    {
        ofile->open();
        /*
         * Let the plugin write a header to the file
         */
        if (plg->active())
        {
            std::string header;
            plg->plugin()->new_file(header);
            if (!header.empty())
                ofile->write(DataFile::NewFile, header.size(), header.c_str());
        }
    }


    if ( plg->active())
        rm->set_plugin(plg->plugin());

    rm->start_of_run();

    daq->nevts( nevts );
    if (plg->active())
    {
        int nn = plg->plugin()->start_of_run(run_type, daq->nevts(), daq->sample_size());
        daq->nevts( nn );
    }
    expected = daq->nevts();
    return 0;
}
int DAQmanager::acquire() {
    ConfigFactory *cfg = ConfigFactory::instance();
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));

    std::ostringstream oheader;
    if (ofile)
        ofile->write(run_type,0,0);

    board->stop(false);
    oheader << "V2.0|";

    running(true);
    stop(false);

    int rc = rm->loop(oheader);
    if ( rc >= 0 )
    {
        if ( plg->active() )
        {
            std::string buffer;
            plg->plugin()->end_of_run(buffer);
            if ( ofile )
            {
                if (!buffer.empty())
                    ofile->write(DataFile::EndOfRun, buffer.size(), buffer.c_str());
            }
        }
        if (ofile)
        {
            std::ostringstream ostr;
            ofile->set_header(oheader.str());
            ofile->write(run_type, ana->pedestals(), ana->noise());
            ofile->close();

            write_msg(_("Data stored in %s\n"), ofile->file_name().c_str());
            delete ofile;
            ofile = 0;
        }
        write_msg(_("Run finished\n"));

    }
    else
    {
        reset();
    }
    running(false);
    stop(false);
    return rc;
}

int DAQmanager::stop_run()
{
    stop(true);
    return 0;
}


void DAQmanager::receive(const EventData *b)
{
    received++;
    process_event(b);
}
void DAQmanager::process_event(const EventData *b)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));
    ana->process_event((unsigned short *)b->chip(0), board->mask());
    if (plg->active())
    {
        std::string data;
        plg->plugin()->filter_event(b, data);
        if (ofile && !data.empty())
        {
            ofile->write(DataFile::DataBlock, data.size(), data.c_str());
        }
    }
    else {
        // write data
        if (ofile)
            ofile->write(b);
    }
}


void DAQmanager::load_config(const char *fnam)
{
    ConfigFactory *cfg = ConfigFactory::instance();

    if (config_file)
        delete config_file;

    config_file = new KeyFile();

    std::ifstream ifile(fnam);
    if (!ifile)
    {
        err_msg( _("could not open config file %s"), fnam);
        return;
    }
    config_file->load(ifile);
    ifile.close();
    cfg->set_file(config_file);
    cfg->config_board( *get_board() );

    config_loaded();
}

void DAQmanager::config_loaded()
{

}

void DAQmanager::save_config(const char *fnam)
{
    ConfigFactory::instance()->save_data();

    std::ofstream ofile(fnam);
    if (!ofile)
    {
        err_msg( _("could not dump config: %s"), fnam );
    }
    else
    {
        config_file->save(ofile);
    }
}


