/*
 * PluginConfig.cc
 *
 *  Created on: Jul 23, 2009
 *      Author: lacasta
 */
#include <glib.h>
#if HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#endif


#include <iostream>
#include <string>
#include <cstdlib>
#include "PluginConfig.h"
#include "SearchPath.h"
#include "string_utils.h"
#include "utils.h"

#ifdef G_OS_UNIX
#define PYTHON_LIBRARY "libAlibavaPython.so"
#elif defined(G_OS_WIN32)
#define PYTHON_LIBRARY "libAlibavaPython.dll"
#endif

class PythonModule
{
    private:
        GModule *module;
        void  (*_unload_module)(gpointer);
        void (*_add_path)(const gchar *);
        void (*_show_path)();
        Plugin *(*_load_plugin)(const gchar *, const gchar *);
        void (*_find_symbols)(const gchar *, std::vector< std::string > &);

    public:
        PythonModule();
        ~PythonModule();

        void unload_module(gpointer M)
        {
            if (_unload_module)
                _unload_module(M);
        }
        void add_path(const char *p)
        {
            if (_add_path)
                _add_path(p);
        }
        Plugin *load_plugin(const char *module, const char *obj)
        {
            if (_load_plugin)
                return _load_plugin(module, obj);
            else
                return 0;
        }
        void show_path()
        {
            if (_show_path)
                _show_path();
        }

        void find_symbols(const gchar *path, std::vector< std::string > &symb)
        {
            if (_find_symbols)
                _find_symbols(path, symb);
        }
};
PythonModule::PythonModule()
    : module(0),
      _unload_module(0), _add_path(0), _show_path(0), _load_plugin(0)
{
    module = g_module_open(PYTHON_LIBRARY, (GModuleFlags)G_MODULE_BIND_LAZY);
    if (!module)
    {
        write_msg(_("Python is not available in your system\n\t%s\n"), g_module_error());
        return;
    }
    bool has_error = false;
    // Now for the functions
    if (!g_module_symbol(module, "unload_module", (gpointer *)&_unload_module))
    {
        err_msg(_("Error loading Python: unload_module: %s"), g_module_error());
        has_error = true;
    }
    if (!g_module_symbol(module, "add_path", (gpointer *)&_add_path))
    {
        has_error = true;
        err_msg(_("Error loading Python: add_path : %s"), g_module_error());
    }
    if (!g_module_symbol(module, "load_plugin", (gpointer *)&_load_plugin))
    {
        has_error = true;
        err_msg(_("Error loading Python: load_plugin : %s"), g_module_error());
    }
    if (!g_module_symbol(module, "show_path", (gpointer *)&_show_path))
    {
        has_error = true;
        err_msg(_("Error loading Python: show_path : %s"), g_module_error());
    }
    if (!g_module_symbol(module, "find_symbols_in_module", (gpointer *)&_find_symbols))
    {
        has_error = true;
        err_msg(_("Error loading Python Wrapper: %s"), g_module_error());
    }

    if (has_error)
    {
        _unload_module = 0;
        _add_path = 0;
        _show_path = 0;
        _load_plugin = 0;
        _find_symbols = 0;
        if (module)
            g_module_close(module);

        module = 0;
    }
}
PythonModule::~PythonModule()
{
    if (module)
        g_module_close(module);
}



static PythonModule _python;

void find_python_symbols(const gchar *module, std::vector< std::string> &symb)
{
    _python.find_symbols(module, symb);
}

struct PythonFunctions
{
        GModule *_lib;
};

struct PluginPriv
{
        Plugin *_plugin;
        GModule *_module;
        gpointer *_pymodule;
};



PluginConfig::PluginConfig() :
    ConfigFile("plugin"), _lib(""), _hook(""), _lang(""), priv(0), _active(false)
{
    init();
}

PluginConfig::PluginConfig(const std::string &lib, const std::string &hook, const std::string &lang) :
    ConfigFile("plugin"),_lib(lib), _hook(hook), _lang(lang), priv(0), _active(false)
{
    init();
}

PluginConfig::PluginConfig(const PluginConfig &P)
    : priv(0)
{
    init();
    cpy(P);
}

void PluginConfig::cpy(const PluginConfig &P)
{
    unload_lib();
    _lib = P.lib();
    _path = P.path();
    _hook = P.hook();
    _lang = P.lang();
    _active = P.active();

    load_lib();
}

void PluginConfig::init()
{
    if (!priv)
        priv = g_new0(PluginPriv, 1);

    if (!priv->_plugin)
        priv->_plugin = new Plugin();



}

PluginConfig::~PluginConfig()
{
    unload_lib();
    g_free(priv);
}

void PluginConfig::active(const bool xxx)
{
    _active = xxx;
    priv->_plugin->active(xxx);
}

Plugin *PluginConfig::plugin()
{
    return priv->_plugin;
}

char *PluginConfig::lib_path()
{
    gchar *filename = g_build_filename(path().c_str(), lib().c_str(),NULL);
    return filename;
}

void PluginConfig::unload_lib()
{
    if (priv->_plugin)
    {
        delete priv->_plugin;
        priv->_plugin = 0;
    }
    if (priv->_module)
    {
        g_module_close(priv->_module);
        priv->_module = 0;
    }
    if (priv->_pymodule)
    {
        _python.unload_module(priv->_pymodule);
        priv->_pymodule = 0;
    }
}

void PluginConfig::load_lib()
{
    SearchPath P;
    std::string fnam;

    if (!_path.empty() )
    {
        P.add_path(_path);
        SearchPath::iterator ip;
        if (_lang == "python")
        {
            for (ip=P.begin(); ip!=P.end(); ++ip)
            {
                const char *pp = ip->c_str();
                _python.add_path(pp);
            }
        }
        else
        {
            const char *ldpath = getenv("LD_LIBRARY_PATH");
            if (ldpath)
                P.add_path(ldpath);

            if (!_lib.empty())
                fnam = P.find_file(_lib);
        }
    }
    if (  ! (_lib.empty() || _hook.empty()) )
    {
        if ( _lang == "python" )
        {
            // python
            priv->_plugin = _python.load_plugin(_lib.c_str(), _hook.c_str());
            if (!priv->_plugin)
            {
                priv->_plugin = new Plugin();
            }
            else if (!priv->_plugin->good())
            {
                delete priv->_plugin;
                priv->_plugin = new Plugin();
                if ( priv->_pymodule )
                {
                    _python.unload_module(priv->_pymodule);
                    _python.show_path();
                }
            }
        }
        else
        {
            // C++
            if (!fnam.empty())
            {
                priv->_module = g_module_open(fnam.c_str(), (GModuleFlags)G_MODULE_BIND_LAZY);
                if (!priv->_module)
                {
                    err_msg(_("Could not load library %s\n"), _lib.c_str());
                }
                else
                {
                    PluginBuilder func;
                    if (!g_module_symbol(priv->_module, _hook.c_str(), (gpointer *)&func) || !func)
                    {
                        err_msg(_("Error loading symbol %s from %s\n"), _hook.c_str(), _lib.c_str());
                    }
                    else
                    {
                        if (priv->_plugin)
                            delete priv->_plugin;
                        priv->_plugin = func();
                    }
                }
            }
        }
    }
}

void PluginConfig::load_data()
{
    if (exists() )
    {
        _lib = get_string("lib");
        _hook = get_string("hook");
        _lang = lower_string( get_string("lang") );
        _path = get_string("path");
        _active = get_bool("active_on_load", false);
        load_lib();
    }
}

void PluginConfig::save_data()
{
    set_string("lib", _lib);
    set_string("hook", _hook);
    set_string("lang", _lang);
    set_string("path", _path);
    set_bool("active_on_load", _active);
}
