/*
 * alibavaClient.cc
 *
 *  Created on: Jan 21, 2014
 *      Author: lacasta
 */
#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>

#include "alibavaClient.h"

std::ostream &operator<<(std::ostream &os, const ns1__Status &st)
{
    double rate = st.rate;
    std::string rate_units = " Hz";
    std::string thr_units;

    if (rate > 1.e6)
    {
        rate /= 1.e6;
        rate_units = "MHz";
    }
    else if (rate > 1000.)
    {
        rate /= 1000.;
        rate_units = "kHz";
    }

    os << "Evts " << std::setiosflags(std::ios::fixed | std::ios::right)
       << std::setw(7) << st.ntrigger << " time " << std::setw(6)
       << std::setprecision(1) << st.time << " rate " << std::setw(8)
       << std::setprecision(1) << rate << rate_units << " eff " << std::setw(5)
       << "[" << st.state << "]"
       << std::flush;
    return os;
}

alibavaClient::alibavaClient()
: server(0)
{

}

void alibavaClient::connect(const char *uri)
{
    if (server)
        delete server;

    server = new alibavaSOAPProxy(uri);
}
alibavaClient::~alibavaClient()
{
    if (server)
        delete server;

    reset_histograms();
}

struct HstEraser : public std::unary_function<Histogram *, void>
{
        HstEraser() {}
        void operator() (Histogram *hst) {
            delete hst;
        }
};
void alibavaClient::reset_histograms()
{
    std::for_each(histograms.begin(), histograms.end(), HstEraser() );
    histograms.clear();
}

int alibavaClient::setDataFile(const char *ofile)
{
    if (!server)
        return 1;

    ns1__setDataFileResponse response;
    return server->setDataFile(ofile, response);
}

int alibavaClient::getStatus(ns1__Status &status)
{
    if (!server)
        return 1;
    ns1__getStatusResponse response;
    server->getStatus(0, response);

    ns1__Status *ss = response.status;
    status.state = ss->state;
    status.nexpected = ss->nexpected;
    status.ntrigger = ss->ntrigger;
    status.rate = ss->rate;
    status.time = ss->time;
    status.value = ss->value;

    return 0;
}

int alibavaClient::stopRun()
{
    if (!server)
        return 1;

    ns1__stopRunResponse response;
    return server->stopRun(0, response);
}



int alibavaClient::startRun(ns1__RunType type, bool async, int nevt, int nbin, int xmin, int xmax, int nsample)
{
    ns1__startRunResponse response;
    ns1__DAQParam param;
    int rc = 0;

    if (!server)
        return 1;

    param.nbin = nbin;
    param.xmin = xmin;
    param.xmax = xmax;
    param.numEvents = nevt;
    param.sampleSize = nsample;
    param.runType = type;

    if (!async)
    {
        server->startRun(&param, response);

        std::istringstream is( std::string((const char *)response.startRunResponse.__ptr,
                                           (size_t)response.startRunResponse.__size ) );




        int ihst, nhst;
        const char *titles[] = { "Spectrum", "Channel Profile", 0 };
        is.read((char *)&nhst, sizeof(int));
        for(ihst=0; ihst<nhst;ihst++)
        {
            Histogram *hst = new Histogram(titles[ihst % 2], titles[ihst % 2], 1, 0., 1.);
            hst->import_from_soap(is);

            histograms.push_back(hst);
        }
    }
    else
        rc = server->startRunAsync(&param);

    return rc;
}

int alibavaClient::startPedestalRun(int nevt, bool async, int nbin, int xmin, int xmax, int nsample)
{

    return startRun(ns1__RunType__Pedestal, async, nevt, nbin, xmin, xmax, nsample);

}


int alibavaClient::startSourceRun(int nevt, bool async, int nbin, int xmin, int xmax, int nsample)
{
    return startRun(ns1__RunType__RadioactiveSource, async, nevt, nbin, xmin, xmax, nsample);
}


int alibavaClient::startLaserRun(int nevt, bool async, int nbin, int xmin, int xmax, int nsample)
{
    return startRun(ns1__RunType__Laser, async, nevt, nbin, xmin, xmax, nsample);
}

